/*
 * sendsms.c
 *
 *  Created on: 03.08.2011
 *      Author: Sergey Kolotsey
 *      E-mail: <kolotsey@gmail.com>
 */

#include <stdlib.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdint.h>
#include <inttypes.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <sys/socket.h>
#include <time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <errno.h>


#ifndef max
#define max(a, b) ({\
	int _a=(a), _b=(b);\
	_a>_b? _a : _b;\
})
#endif

#ifndef min
#define min(a, b) ({\
	int _a=(a), _b=(b);\
	_a<_b? _a : _b;\
})
#endif

static char *_program = "sendsms";

/**
 * This function prints usage message to output
 *
 * @param f output stream: either stderr or stdout
 */
void usage(FILE *f) {
	fprintf(f,
			"%s - send SMS message via Sky Control monitoring system\n"
			"\n"
			"Usage: %s [-p PASSWORD] HOST PHONE [MESSAGE]\n"
			"Where:\n"
			"	HOST     - System's IP address or host name\n"
			"	PHONE    - Phone number to send SMS to\n"
			"	MESSAGE  - SMS message to send\n"
			"	PASSWORD - System password if required\n", _program, _program);
}

/**
 * Trims buffer and returns pointer to the beginning of the trimmed string
 * This function modifies original buffer
 *
 * @param s Pointer to buffer containing a string to trim
 * @return Pointer to buffer with trimmed string
 */
static char *trim(char *s) {
	char *ps = s;
	//Trims: 0x09 (tab), 0x20 (space), 0x0a (new line), 0x0d (carrier return)
	while ((ps[0] != 0x00) && (ps[0] == 0x09 || ps[0] == 0x20 || ps[0] == 0x0a || ps[0] == 0x0d)) ps++;
	while (strlen(ps) && (ps[strlen(ps) - 1] == 0x09 || ps[strlen(ps) - 1] == 0x20 || ps[strlen(ps) - 1] == 0x0a || ps[strlen(ps) - 1] == 0x0d)) ps[strlen(ps) - 1] = 0x00;
	return ps;
}

/**
 * Connect to remote server to specified port using TCP
 *
 * @param host IP address or hostname of remote node
 * @param port Remote port number
 * @return Socket file descriptor if connection succeded. If connection failed,
 *    -1 is returned and errno is set
 */
int connect_tcp(char *host, int port) {
	int fd;
	struct sockaddr_in sin;
	struct hostent *he;

	if ((he = gethostbyname(host)) == NULL) {
		errno = EHOSTUNREACH;
		return -1;
	}

	if (-1 == (fd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP))) {
		return -1;

	} else {
		memset(&sin, 0, sizeof(struct sockaddr_in));
		sin.sin_family = AF_INET;
		sin.sin_port = htons(port);
		memcpy((char *) &sin.sin_addr, he->h_addr, he->h_length);

		if (-1 == connect(fd, (struct sockaddr *) &sin,
				sizeof(struct sockaddr_in))) {
			close(fd);
			return -1;
		} else {
			return fd;
		}
	}
}

/**
 * Given an opened socket, writes data from buffer to this socket and expects
 * for response from remote server. This response is then written to the same
 * buffer and its length is returned.
 *
 * @param sfd Socket to write data to
 * @param timeout Timeout in seconds for expectation of reply
 * @param buffer Pointer to buffer where data to be sent is stored and where
 *    function can store received data
 * @param buffer_len Length of data to send
 * @param buffer_maxlen Maximum length of buffer
 * @return Length of received data or -1 if error occured (errno is set)
 */
int make_request(int sfd, int timeout, char *buffer, int buffer_len,
		int buffer_maxlen) {
	struct timeval tv;
	fd_set rfd;
	int n_total = 0;
	int n;

	//Write request to socket
	if (-1 == write(sfd, buffer, strlen(buffer))) {
		return -1;
	}

	//Expect data from remote host
	while (1) {
		FD_ZERO( &rfd);
		FD_SET( sfd, &rfd);

		if (timeout == 0) {
			n = select(sfd + 1, &rfd, 0, 0, NULL);
		} else {
			tv.tv_sec = timeout;
			tv.tv_usec = 0;
			n = select(sfd + 1, &rfd, 0, 0, &tv);
		}

		switch (n) {
		case -1:
			return -1;
			break;

		case 0:
			errno = ETIMEDOUT;
			return -1;
			break;

		default:
			if (-1 == (n = read(sfd, buffer + n_total,
					buffer_maxlen - n_total))) {
				return -1;

			} else if (0 == n) {
				return n_total;

			} else {
				n_total += n;
			}
			break;
		}
	}
	return 0;
}

/**
 * This function gets buffer with string that was received from remote system
 * and parses it and examines if the system returned any valid status code.
 *
 * @param response Pointer to buffer containing response string with
 *    HTTP headers and status code
 * @return 0 if response code is positive (no error), 1 otherwise
 */
int parse_response(char *response) {
	char *p = response;
	char *header;
	char *tok;
	int hi=0;

	//In the response there are HTTP headers in the beginning.
	//We need to remove them and in the same time check if HTTP
	//status code is `200 OK'
	header = p;
	while (*p) {
		if ((*p == '\r' && hi == 0) ||
		        (*p == '\n' && hi == 1) ||
		        (*p == '\r' && hi == 2) ||
		        (*p == '\n' && hi == 3)) {
			hi++;
		} else {
			hi = 0;
		}

		if (hi == 2) {
			//End of next header line found
			//Trim line and check it
			*p = 0;
			header = trim(header);
			if (0 == strncasecmp(header, "HTTP", 4)) {
				int status;
				if ((tok = strchr(header, ' '))) {
					header = trim(tok);
					status = atoi(header);
					if (status != 200) {
						fprintf( stderr,
								"Remote system returned error status: %s\n"
								"Your message was not sent\n",
								header);
						return -1;
					}
				} else {
					fprintf(stderr, "Remote system returned invalid HTTP header: %s\n",
							header);
					return 1;
				}
			}
			p++;
			header = p;
		} else {
			p++;
		}

		if (hi == 4) {
			//End of headers found
			break;
		}
	}

	//After the headers are removed, check response string.
	//Monitoring system returns response as XML string in the following
	//format:
	//	<error type="no error"/>
	//if no error occured. Error type may be "access error" if system wants
	//user to specify password to send SMS
	response = trim(p);
	if ((tok = strstr(response, "type="))) {
		response = tok + 5;
		char quot = *response;
		response += 1;
		if ((tok = strchr(response, quot))) {
			*tok = 0;
			if (0 == strcasecmp(response, "no error")) {
				fprintf(stderr, "Your message has been sent\n");
				return 0;

			} else if (0 == strcasecmp(response, "access error")) {
				fprintf(
						stderr,
						"Could not send SMS message without sending password.\n"
						"Specify password with -p argument\n"
						"Your message was not sent\n");
				return 1;

			} else {
				fprintf(stderr,
						"Remote system returned error code: %s\n"
						"Your message was not sent\n",
						response);
				return 1;
			}

		} else {
			fprintf(stderr,
					"Could not understand response from remote system\n");
			return -1;
		}

	} else {
		fprintf(stderr,
				"Could not understand response from remote system: %s\n",
				response);
		return 1;
	}
}


int main(int argc, char *argv[]) {
	int sfd;                  //Socket file descriptor.
	char *password = NULL,    //Parameters that user specifies in
			*address, *phone; //command line agruments
	char message[BUFSIZ];     //Message that has to be sent will be stored here
	int c, i;

	if (argc) {
		_program = argv[0];
	}

	//Check if user wants to see help message
	for (i = 1; i < argc; i++) {
		if (0 == strcmp("--help", argv[i])) {
			usage(stdout);
			return 0;
		}
	}

	//Parse command line arguments/flags
	while ((c = getopt(argc, argv, "hp:")) != -1) {
		switch (c) {
		case 'h':
			usage(stdout);
			return 0;
			break;
		case 'p':
			password = optarg;
			break;
		default:
			break;
		}
	}
	argc -= optind;
	argv += optind;

	//Check if there are a correct number of arguments provided
	if (argc < 2) {
		usage(stderr);
		return 1;
	}

	//Get system address, phone number and message from cmd args
	address = argv[0];
	phone = argv[1];
	strcpy(message, "");
	for (i = 2; i < argc; i++) {
		if (i){
			strncat(message, " ", BUFSIZ - strlen(message));
		}
		strncat(message, argv[i], BUFSIZ - strlen(message));
	}

	//Connect to remote system to port 80 (HTTP)
	sfd = connect_tcp(address, 80);
	if (-1 == sfd) {
		fprintf(stderr, "Could not connect to remote host: %s\n",
				strerror(errno));
		return 1;

	} else {
		char query[BUFSIZ];
		char command[BUFSIZ];
		int saved_errno = errno;
		int ret;

		//When connection is established, form a request using parameters that
		//user specified in command line and send the request to remote system
		if (password) {
			snprintf(command, BUFSIZ,
					"epassword=%s\r\n"
					"querytype=send_sms_message\r\n"
					"phone=%s\r\n"
					"message=%s",
					password, phone, message);
		} else {
			snprintf(command, BUFSIZ,
					"querytype=send_sms_message\r\n"
					"phone=%s\r\n"
					"message=%s",
					phone, message);
		}
		command[BUFSIZ - 1] = 0;

		//The request is a HTTP POST request
		snprintf(query, BUFSIZ, "POST /engine.htm?random=%ld HTTP/1.0\r\n"
				"Host: %s\r\n"
				"User-Agent: Mozilla/4.0\r\n"
				"Connection: close\r\n"
				"Content-type: application/x-www-form-urlencoded\r\n"
				"Content-length: %ld\r\n"
				"Pragma: no-cache\r\n"
				"Accept: */*\r\n"
				"\r\n"
				"%s", time(NULL), address, (long int) strlen(command), command);
		query[BUFSIZ - 1] = 0;

		//Send request to remote system and get response
		//to the same buffer `query'
		if (-1 == (ret = make_request(sfd, 5, query, strlen(query), BUFSIZ))) {
			fprintf(stderr, "Could not communicate with remote system: %s\n",
					strerror(saved_errno));
		} else {
			query[ret] = 0;
			//Now parse the response that system returned
			//and remember status code
			ret = parse_response(query);
		}
		shutdown(sfd, SHUT_RDWR);
		close(sfd);
		return ret;
	}
}

